# Neural networks from scratch with numpy.

import numpy as np
import pygrad.tensor as tensor

def mean_absolute_error(x, y):
  return np.mean(np.abs(x - y))

def mean_squared_error(x, y):
  return np.mean(np.power(x - y, 2))

def cross_entropy_loss(x, y):
  return -np.log(np.exp(y) / np.sum(np.exp(x)))

# prepare inputs and outputs
x = np.array([[1, 0]])
y = np.array([[1]])

# we're doing xavier initialisation - see <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>
w1 = np.random.randn(2, 3) / np.sqrt(2)
w2 = np.random.randn(3, 1) / np.sqrt(3)

def single_pass():
  # forward pass
  h = np.matmul(x, w1)
  h_hat = np.tanh(h)
  j = np.matmul(h_hat, w2)
  print("prediction {}".format(j))

  # loss calculation
  loss = cross_entropy_loss(j, y)
  print("loss {}".format(loss))

  # TODO Backward pass.
  return

# initialise layers
#   self.lin1 = nn.Linear(2, 3)
#   self.lin2 = nn.Linear(3, 1)
#   self.loss = nn.MSELoss()
# and then
#   x = self.lin1(x)
#   x = F.relu(x)
#   x = self.lin2(x)
#   x = F.softmax(x)
#   loss = self.loss(x, y)

# Small test to see if autograd works.
def test():
  # Input tensors.
  x, y, z = tensor.Tensor(-2), tensor.Tensor(5), tensor.Tensor(-4)
  
  # Forward pass.
  q = x.add(y)
  h = q.expt(2)
  w = h.mul(z)
  print(f"q = {q}, w = {w}")

  # Backward pass.
  w.backward()
  print(f"is: dw = {w.grad}, dz = {z.grad}, dy = {y.grad}, dx = {x.grad}")
